/*
 *  nfs/nfs4proc.c
 *
 *  nfs version 4 client side proceedure stub
 *
 *  by Andy Adamson <andros@citi.umich.edu>
 *  and Jim Rees <rees@umich.edu>
 *  Copyright (C) 1999, 2000 The Regents of the University of Michigan
 *
 *  $Id: nfs4proc.c,v 1.8 2000/07/17 17:33:29 rees Exp $
 */


#define NFS_NEED_NFS4_XDR_TYPES

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/proc.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#include <sys/kernel.h>
#include <sys/mount.h>
#include <sys/buf.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/systm.h>

#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>

#include <nfs/rpcv2.h>
#include <nfs/nfsproto.h>
#include <nfs/nfsnode.h>
#include <nfs/nfs.h>
#include <nfs/nfsmount.h>
#include <nfs/xdr_subs.h>
#include <nfs/nfsm_subs.h>
#include <nfs/nfsdiskless.h>
#include <nfs/nqnfs.h>
#include <nfs/nfs_var.h>
#include <nfs/nfs4_prot.h>

/* Global State */

clientid4 nfs4_clientid = 0; 			/* set at mount */
verifier4 nfs_verf;				/* set at init */

int
nfs4_proc_compound(struct vnode *vp, struct proc *procp, struct ucred *cred, COMPOUND4args *cargp, COMPOUND4res *cresp)
{
    u32 sendsize, recvsize;
    caddr_t bpos, dpos;
    int error;
    struct mbuf *mreq, *mrep, *md, *mb, *mb2;
    struct rpc_rqst req;
    struct iovec iov;

    nfsv4_printk(compound,"NFSV4 call compound\n");

    error = nfs4_get_sizes(&sendsize, &recvsize, cargp, cresp);
    if (error != 0)
	goto nfsmout;
    nfsm4_reqhead(sendsize);
    nfsm4_xdrargs(cargp);

    nfsm4_request(vp, procp, cred);

    error = nfs4_xdr_compoundres(&req, (u32 *) dpos, cresp);

    m_freem(mrep);

 nfsmout:
    nfsv4_printk(compound,"NFSV4 reply compound. status: %d\n", error);
    return error;
}
